#ifdef HAVE_CONFIG_H
#  include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <stdarg.h>
#include "limits.h" 
#include "cdi.h"
#include "pio_util.h"

#ifdef USE_MPI
#include "pio_interface.h"
#include "pio_comm.h"
#include "pio_rpc.h"
#include "pio_server.h"
#include "namespace.h"
#include "resource_handle.h"
#include "stream_int.h"
#include "vlist.h"
extern resOps streamOps;

size_t         * winBufferSize = NULL;
unsigned char ** winBuffer     = NULL;
unsigned char ** winBufferHead = NULL;
MPI_Win        * win           = NULL;
int            * winPostSet    = NULL;
MPI_Group      * groupsIO      = NULL;
MPI_Group      * groupsIONetto = NULL;

char * funcMap[nFuncs] = {"streamOpen", "streamDefVlist", "streamClose" };


/****************************************************/

int cmp ( const void * va, const void * vb )
{
  const int ** a, ** b;
 
  a = ( const int ** ) va;
  b = ( const int ** ) vb;
  
  return (( **a < **b ) - ( **a > **b ));
}

/****************************************************/
 
void mapProblems ( int problemSizes[], int * problemMapping, int nProblems, 
		   int nWriter, double * w )
{
  int *ip[nProblems];
  int dummy[nProblems];
  int buckets[nWriter];
  int currCapacity, nextCapacity;
  double meanBucket[nWriter];
  int sum = 0;
  int writerIdx = 0;
  int i, j;
  

  for ( i = 0; i < nProblems; i++ )
    {
      ip[i] = &problemSizes[i];
      sum += problemSizes[i];
    }

  qsort ( ip, nProblems, sizeof ( int * ), cmp );

  for ( i = 0; i < nProblems; i++ )
    dummy[i] = ip[i] - problemSizes;
  
  for ( j = 0; j < nWriter; j++ )     
    meanBucket[j] = ( double ) sum * ( * ( w + j ));

  for ( i = 0; i < nProblems; i++ )
    {
      currCapacity = INT_MIN;
      
      for ( j = 0; j < nWriter; j++ )
	{
	  if ( !i ) buckets[j] = 0.0;
	  nextCapacity = meanBucket[j] - ( buckets[j] + ( *ip[i] ));
	
	  if ( nextCapacity > currCapacity )
	    {
	      currCapacity = nextCapacity;
	      writerIdx = j;
	    }
	}
      problemMapping[ dummy[i] ] = writerIdx;
      buckets[writerIdx] +=  *ip[i];
    }

  xprintArray3 (  "problemSizes = ", problemSizes, nProblems, DATATYPE_INT );     
  xprintArray3 ( "vector of indices, qsort of problemSizes", dummy, 
                nProblems, DATATYPE_INT ); 
  xprintArray3 ( "problemMapping", problemMapping, nProblems, DATATYPE_INT );
  xprintArray3 ( "meanBucket", meanBucket, nWriter, DATATYPE_FLT );
  xprintArray3 ( "actual buckets", buckets, nWriter, DATATYPE_INT );      
}

/****************************************************/ 

/**
   @brief is encapsulated in CDI library.

   @param vSizes array with number of levels for all var_t 's in order of streams
   
   @param sSizes array with number of var_t for all stream_t 's

   @param varMapping return value, array with ranks of I/O PEs assigned to var_t 's 
in order of vSizes

   @param nStreams number of stream_t 's
   
   @param nodeSizes array of number of I/O PEs on each physical nodes, increasing 
   order of ranks assumed

   @param nNodes number of physical nodes hosting I/O PEs
   
   @return
*/
  
void  varMapGen ( int * vSizes, int * sSizes, int * varMapping, 
		     int nStreams, int * nodeSizes, int nNodes )
{

  int weightsStreams[nStreams];
  int streamMapping[nStreams];
  int nPEs = 0, nVars = 0;
  int i, j, k, offset = 0, offsetN = 0;
  double * w;
  
  int * weightsVarsNode;
  int * varMappingNode;
  int nVarsNode, summandRank = 0;
  int nProcsColl = commInqNProcsColl ();

  int buckets[nProcsColl];

  for ( i = 0; i < nStreams; i++ )
    {
      nVars += * ( sSizes + i );
      weightsStreams[i] = 0;
      for ( j = 0; j < * ( sSizes + i ); j++ )
	weightsStreams[i] += * ( vSizes + offset++ );
    }

  w = ( double * ) xmalloc ( nNodes * sizeof ( double ));
  for ( j = 0; j < nNodes; j++ )
    nPEs += * ( nodeSizes + j );

  for ( j = 0; j < nNodes; j++ )
    w[j] = ( double ) * ( nodeSizes + j ) / ( double ) nPEs;

  mapProblems ( weightsStreams, streamMapping, nStreams, nNodes, w );

  free ( w );

  for ( i = 0; i < nNodes; i++ )
    {
      nVarsNode = 0;
      for ( j = 0; j < nStreams; j++ )
	if ( * ( streamMapping + j ) == i )
	  nVarsNode += * ( sSizes + j ); 
      
      weightsVarsNode = ( int * ) xmalloc ( nVarsNode * sizeof ( int ));
      varMappingNode = ( int * ) xmalloc ( nVarsNode * sizeof ( int ));
      w = ( double * ) xmalloc ( * ( nodeSizes + i ) * sizeof ( double ));
      offset = 0;
      offsetN = 0;

      for ( j = 0; j < nStreams; j++ )
	if ( * ( streamMapping + j ) == i )
	  for ( k = 0; k < * ( sSizes + j ); k ++ )
	    * ( weightsVarsNode + offsetN++ ) = * ( vSizes + offset++ );
	else
	  offset += * ( sSizes + j );

      for ( j = 0; j < * ( nodeSizes + i ); j ++ )
	w[j] = 1.0 / ( double ) * ( nodeSizes + i );

      mapProblems ( weightsVarsNode, varMappingNode, nVarsNode, 
		    * ( nodeSizes + i ),  w );

      offset = 0;
      offsetN = 0;

      for ( j = 0; j < nStreams; j++ )
	if ( * ( streamMapping + j ) == i )
	  for ( k = 0; k < * ( sSizes + j ); k ++ )
	    * ( varMapping + offset ++ ) = 
              commCollID2RankGlob ( * ( varMappingNode + offsetN ++ ) +  summandRank );
	else
	  offset += * ( sSizes + j );

      summandRank += * ( nodeSizes + i );

      free ( w );
      free ( varMappingNode );
      free ( weightsVarsNode );
    }
  
  if ( ddebug )
    {
      xprintArray ( "varMapping", varMapping, nVars, DATATYPE_INT  );
      for ( i = 0; i < nProcsColl; i++ )
	buckets[i] = 0;
      for ( i = 0; i < nVars; i ++ )
	buckets[commRankGlob2CollID ( *(varMapping + i ))] += * ( vSizes + i );
      xprintArray ( "buckets", buckets, nProcsColl, DATATYPE_INT );
    }
}

/************************************************************************/ 

void defVarDeco ( int vlistID, int varID )
{
  int varSize, cRank, lChunk, rem, lOffset;
  int nProcsModel = commInqNProcsModel ();
  deco_t deco[nProcsModel];
 
  varSize = vlistInqVarSize ( vlistID, varID );

  for ( cRank = 0; cRank < nProcsModel; cRank++ )
    {
      lChunk = varSize / nProcsModel;
      lOffset = cRank * lChunk;
      rem = varSize % nProcsModel;
      if ( cRank < rem ) 
        {
          lChunk++;
          lOffset += cRank;
        }
      else
        lOffset += rem;
      
      deco[cRank].rank   = cRank;
      deco[cRank].offset = lOffset;
      deco[cRank].chunk  = lChunk;
    }

  vlistDefVarDeco ( vlistID, varID, nProcsModel, &deco[0] );
}

/************************************************************************/


static
void varsMapNDeco ( int nNodes, int * nodeSizes )
{
  int nStreams, nVars, * resHs, * streamSizes, * varSizes, * varMapping;
  int i, j, k = 0;

  nStreams = streamSize ();

  resHs       = xmalloc ( nStreams * sizeof ( resHs[0] ));
  streamSizes = xmalloc ( nStreams * sizeof ( streamSizes[0] ));
  streamGetIndexArray ( nStreams, resHs );

  for ( i = 0; i < nStreams; i++ )
    streamSizes[i] = streamInqNvars ( * ( resHs + i )); 

  nVars = xsum ( nStreams, streamSizes );
  varSizes   = xmalloc ( nVars * sizeof ( varSizes[0] ));
  varMapping = xmalloc ( nVars * sizeof ( varMapping[0] ));

  for ( i = 0; i < nStreams; i++ )
    for ( j = 0; j < * ( streamSizes + i ); j++ )
      varSizes[k++] += vlistInqVarSize ( streamInqVlist ( * resHs + i ), j );

  assert ( k == nVars );

  varMapGen ( varSizes, streamSizes, varMapping, 
	      nStreams, nodeSizes, nNodes );

  k = 0;
  for ( i = 0; i < nStreams; i++ )
    for ( j = 0; j < * ( streamSizes + i ); j++ )
      {
        defVarDeco ( streamInqVlist ( * resHs + i ), j );
        defVarDeco ( streamInqVlistIDorig ( * resHs + i ), j );
	vlistDefVarIOrank ( streamInqVlist ( * resHs + i ), j, 
			  * ( varMapping + k ));
	vlistDefVarIOrank ( streamInqVlistIDorig ( * resHs + i ), j, 
			  * ( varMapping + k++ ));
      }

  if ( varMapping )  free ( varMapping );
  if ( varSizes )    free ( varSizes );
  if ( streamSizes ) free ( streamSizes );
  if ( resHs )       free ( resHs );
}

/************************************************************************/

static 
void modelWinCleanup ( void )
{
  int collID;

  for ( collID = 0; collID < commInqNProcsColl (); collID ++ )
    {
      if ( winPostSet[collID] )
        xmpi ( MPI_Win_wait ( win[collID] )); 
      xmpi ( MPI_Win_free ( &win[collID] ));
      xmpi ( MPI_Free_mem ( winBuffer[collID] ));
      xmpi ( MPI_Group_free ( &groupsIONetto[collID] ));
    }

  if ( groupsIONetto )   free ( groupsIONetto );
  if ( winPostSet )      free ( winPostSet );
  if ( win )             free ( win );
  if ( winBufferHead )   free ( winBufferHead );
  if ( winBuffer )       free ( winBuffer );
  if ( winBufferSize )   free ( winBufferSize );

  xdebug ( "cleaned up mpi_win's" );
}

/************************************************************************/

static
  void modelWinDefBufferSizes ( void )
{
  int collID, nstreams, * streamIndexArray, streamNo, vlistID, nvars, varID;
  int collIDchunk = 0, sumWinBufferSize = 0;
  int nProcsColl  = commInqNProcsColl ();
  int rankGlob    = commInqRankGlob ();
  int root = commInqRootGlob ();

  if ( winBufferSize == NULL ) xabort ( "" );
  
  nstreams = reshCountType ( &streamOps );
  streamIndexArray = xmalloc ( nstreams * sizeof ( streamIndexArray[0] ));
  reshGetResHArrayOfType ( nstreams, streamIndexArray, &streamOps );
  for ( streamNo = 0; streamNo < nstreams; streamNo++ )
    {
      // space required for data
      vlistID = streamInqVlist ( streamIndexArray[streamNo] );
      nvars = vlistNvars ( vlistID );
      for ( varID = 0; varID < nvars; varID++ )
        {
          collID = CDI_UNDEFID;
          collID = commRankGlob2CollID ( vlistInqVarIOrank ( vlistID, varID ));
          collIDchunk = vlistInqVarDecoChunk ( vlistID, varID );
          if ( collID == CDI_UNDEFID ||
               collIDchunk <= 0 ) xabort ( "" );
          winBufferSize[collID] += ( collIDchunk * sizeof ( double ) + 
                                     winBufferOverheadChunk  * sizeof ( int ));
        }

      // space required for the 3 function calls streamOpen, streamDefVlist, streamClose 
      // once per stream and timestep for all collprocs only on the modelproc root
      if ( rankGlob == root ) 
        for ( collID = 0; collID < nProcsColl; collID++ ) 
          winBufferSize[collID] += ( 3 * winBufferOverheadFuncCall * sizeof ( int ) +
                                     5 * sizeof ( int ) +
                                     MAXDATAFILENAME );                                     
    }
  free ( streamIndexArray );
  
  for ( collID = 0; collID < nProcsColl; collID++ ) 
    {
      winBufferSize[collID] += ( winBufferOverhead * sizeof ( int )); 
      sumWinBufferSize += winBufferSize[collID];
    }
  if ( sumWinBufferSize > MAXWINBUFFERSIZE ) xabort ( "" );
  xprintArray ( "winBufferSize", winBufferSize, nProcsColl, DATATYPE_INT );
}


/************************************************************************/


static 
  void modelWinFlushBuffer ( int collID )
{
  int nProcsColl = commInqNProcsColl ();

  if ( collID                < 0 ||
       collID                >= nProcsColl ||
       winBuffer             == NULL ||
       winBuffer[collID]     == NULL ||
       winBufferHead         == NULL ||
       winBufferSize         == NULL ||
       winBufferSize[collID] < 0     ||
       winBufferSize[collID] > MAXWINBUFFERSIZE ) 
    xabort ( "" );
  memset ( winBuffer[collID], 0, winBufferSize[collID] );
  winBufferHead[collID] = winBuffer[collID];
}


/************************************************************************/


static
void modelWinCreate ( void )
{
  int collID, ranks[1];
  int nProcsColl = commInqNProcsColl ();

  winBufferSize   = xmalloc ( nProcsColl * sizeof ( winBufferSize[0] ));
  winBuffer       = xmalloc ( nProcsColl * sizeof ( winBuffer[0] ));
  winBufferHead   = xmalloc ( nProcsColl * sizeof ( winBufferHead[0] ));
  win             = xmalloc ( nProcsColl * sizeof ( win[0] ));
  winPostSet      = xmalloc ( nProcsColl * sizeof ( winPostSet[0] ));
  groupsIONetto   = xmalloc ( nProcsColl * sizeof ( groupsIONetto[0] ));

  modelWinDefBufferSizes ();
  ranks[0] = commInqNProcsModel ();

  for ( collID = 0; collID < nProcsColl; collID ++ )
    {
      if ( winBufferSize[collID] <= 0 ) xabort ( "" );
      winBuffer[collID] = NULL;
      xmpi ( MPI_Alloc_mem (( MPI_Aint ) winBufferSize[collID], 
                            MPI_INFO_NULL, &winBuffer[collID] ));
      if ( winBuffer[collID] == NULL ) xabort ( "" ); 
      winBufferHead[collID] = winBuffer[collID];
      xmpi ( MPI_Win_create ( winBuffer[collID],  
                              ( MPI_Aint ) winBufferSize[collID], 1,
                              MPI_INFO_NULL, commInqCommsIO ( collID ), 
                              &win[collID] )); 
      xmpi ( MPI_Comm_group ( commInqCommsIO ( collID ), &groupsIONetto[collID] ));
      xmpi ( MPI_Group_incl ( groupsIONetto[collID], 1, ranks, 
                              &groupsIONetto[collID] ));
    }
  xdebug ( "created mpi_win's" );
}

/************************************************************************/

static
  void modelWinBufferPutAtEnd ( const char * caller, 
                                int ID, void * argBuffer, size_t size )
{
  char text[1024];
  /*
    xdebug ( "ID=%d, size=%d, newBufferHead=%d, oldBufferSize=%d",
    ID, size, winBufferHead[ID] - winBuffer[ID] + size, winBufferSize[ID] );
  */
  if ( winBufferHead == NULL ||
       argBuffer     == NULL ||
       size           < 0    ||
       ID             < 0    ||
       ID             >= commInqNProcsColl () ||
       winBufferHead[ID] - winBuffer[ID] + size > winBufferSize[ID] )
    {
      sprintf ( text, "caller: %s", caller );
      xabort ( text );
    }
  
  memcpy ( winBufferHead[ID], argBuffer, size );
  winBufferHead[ID] += size;
}

/************************************************************************/

void pioBufferData ( int streamIDarg, int varIDarg, const double *dataArg, int nmiss )
{
  int chunk, streamID, varID, vlistID, collID = CDI_UNDEFID;
  int tokenSep = SEPARATOR, tokenData = DATATOKEN;
  size_t size;
  void * vp;
  double * data;

  data     = dataArg;
  streamID = streamIDarg;
  varID    = varIDarg;
  vlistID  = streamInqVlist ( streamID );
  collID   = commRankGlob2CollID ( vlistInqVarIOrank ( vlistID, varID ));
  chunk    = vlistInqVarDecoChunk ( vlistID, varID );
  if ( collID         < 0                    ||
       collID        >= commInqNProcsColl () ||
       chunk          < 0                    ||
       winBufferSize == NULL                 ||
       winBuffer     == NULL                 ||
       winBufferHead == NULL                 ||
       winPostSet    == NULL                 ||
       win           == NULL )
    xabort ( "" );
  
  xdebug3 ("streamID=%d, vlistID=%d, varID=%d, collID=%d, "
          "chunk=%d, data=%.2f, nmiss=%d",
          streamID, vlistID, varID, collID, chunk, data[0], nmiss );  

  if ( winPostSet[collID] ) 
    { 
      xmpi ( MPI_Win_wait ( win[collID] )); 
      winPostSet[collID] = 0;
      modelWinFlushBuffer ( collID );
    }

  size = chunk * sizeof ( double ) + winBufferOverheadChunk * sizeof ( int );
  if ( winBufferHead[collID] - winBuffer[collID] + size >= winBufferSize[collID] )
    xabort ( "" ); 

  modelWinBufferPutAtEnd ( __func__, collID, &tokenData, sizeof ( tokenData ));
  modelWinBufferPutAtEnd ( __func__, collID, &streamID , sizeof ( streamID ));
  modelWinBufferPutAtEnd ( __func__, collID, &varID    , sizeof ( varID ));
  modelWinBufferPutAtEnd ( __func__, collID, data      , chunk * sizeof ( double ));
  modelWinBufferPutAtEnd ( __func__, collID, &tokenSep , sizeof ( tokenSep ));      
}

/************************************************************************/

int func2Int ( const char * func )
{
  int iret = CDI_UNDEFID, funcID;

  for ( funcID = 0; funcID < nFuncs; funcID ++ )
    if ( strlen ( funcMap[funcID] ) == strlen ( func ) && 
         memcmp ( func, funcMap[funcID], strlen ( func )) == 0 )
      iret = funcID;

  return iret;
}

/************************************************************************/

void pioBufferFuncCall ( const char * func, int argc, ... )
{
  va_list ap;
  int rankGlob = commInqRankGlob ();
  int root = commInqRootGlob ();
  int collID, nProcsColl = commInqNProcsColl ();
  int tokenSep = SEPARATOR, tokenFuncCall = FUNCCALL;
  size_t size = 0;
  int funcID = func2Int ( func );
  
  if ( rankGlob != root ) return;

  if ( func          == NULL                 ||
       argc           < 1                    ||
       argc           > 2                    ||
       winBufferSize == NULL                 ||
       winBuffer     == NULL                 ||
       winBufferHead == NULL                 ||
       winPostSet    == NULL                 ||
       win           == NULL )
    xabort ( "" );
         
  va_start ( ap, argc );
  
  switch ( funcID ) 
    {
    case STREAMCLOSE:
      {
        int streamID;

        if ( argc != 1 ) xabort ( "" );
        streamID  = va_arg ( ap, int );
        
        for ( collID = 0; collID < nProcsColl; collID++ ) 
          {
            size = ( winBufferOverheadFuncCall + 1 ) * sizeof ( int );
            if ( winBufferHead[collID] - winBuffer[collID] + size >= 
                 winBufferSize[collID] )
              xabort ( "" );  

            if ( winPostSet[collID] ) 
              { 
                xmpi ( MPI_Win_wait ( win[collID] )); 
                winPostSet[collID] = 0;
                modelWinFlushBuffer ( collID );
              }
            modelWinBufferPutAtEnd ( __func__, collID, &tokenFuncCall, 
                                     sizeof ( tokenFuncCall));
            modelWinBufferPutAtEnd ( __func__, collID, &funcID, 
                                     sizeof ( funcID ));
            modelWinBufferPutAtEnd ( __func__, collID, &streamID, 
                                     sizeof ( streamID ));
            modelWinBufferPutAtEnd ( __func__, collID, &tokenSep, 
                                     sizeof ( tokenSep ));
          }
      xdebug ( "WROTE FUNCTION CALL IN BUFFER OF WINS:  %s, streamID=%d", 
               funcMap[funcID], streamID );
      }       
      break;
    case STREAMOPEN:
      {
        char * filename;
        int    filetype;
        size_t filenamesz;
        
        if ( argc != 2 ) xabort ( "" );
        filename  = va_arg ( ap, char * );
        filenamesz = strlen ( filename );
        if ( filenamesz <= 0 &&
             filenamesz >= MAXDATAFILENAME ) xabort ( "" );
        filetype  = va_arg ( ap, int );
        
        for ( collID = 0; collID < nProcsColl; collID++ ) 
          {
            size = ( winBufferOverheadFuncCall + 2 ) * sizeof ( int ) +  
              MAXDATAFILENAME;
            if ( winBufferHead[collID] - winBuffer[collID] + size >= 
                 winBufferSize[collID] )
              xabort ( "" );  

            if ( winPostSet[collID] ) 
              { 
                xmpi ( MPI_Win_wait ( win[collID] )); 
                winPostSet[collID] = 0;
                modelWinFlushBuffer ( collID );
              }
            modelWinBufferPutAtEnd ( __func__, collID, &tokenFuncCall, 
                                     sizeof ( tokenFuncCall));
            modelWinBufferPutAtEnd ( __func__, collID, &funcID, 
                                     sizeof ( funcID ));
            modelWinBufferPutAtEnd ( __func__, collID, &filenamesz, 
                                     sizeof ( filenamesz ));
            modelWinBufferPutAtEnd ( __func__, collID, filename, 
                                     filenamesz );
            modelWinBufferPutAtEnd ( __func__, collID, &filetype, 
                                     sizeof ( filetype ));
            modelWinBufferPutAtEnd ( __func__, collID, &tokenSep, 
                                     sizeof ( tokenSep ));
          }
        
        xdebug ( "WROTE FUNCTION CALL IN BUFFER OF WINS:  %s, filenamesz=%d,"
                 " filename=%s, filetype=%d", 
                 funcMap[funcID], filenamesz, filename, filetype );
      }
      break;
    case STREAMDEFVLIST:
      {
        int streamID, vlistID;

        if ( argc != 2 ) xabort ( "" );
        streamID  = va_arg ( ap, int );
        vlistID   = va_arg ( ap, int );
        
        for ( collID = 0; collID < nProcsColl; collID++ ) 
          {
            size = ( winBufferOverheadFuncCall + 2 ) * sizeof ( int );
            if ( winBufferHead[collID] - winBuffer[collID] + size >= 
                 winBufferSize[collID] )
              xabort ( "" );  

            if ( winPostSet[collID] ) 
              { 
                xmpi ( MPI_Win_wait ( win[collID] )); 
                winPostSet[collID] = 0;
                modelWinFlushBuffer ( collID );
              }
            modelWinBufferPutAtEnd ( __func__, collID, &tokenFuncCall, 
                                     sizeof ( tokenFuncCall));
            modelWinBufferPutAtEnd ( __func__, collID, &funcID, 
                                     sizeof ( funcID ));
            modelWinBufferPutAtEnd ( __func__, collID, &streamID, 
                                     sizeof ( streamID ));
            modelWinBufferPutAtEnd ( __func__, collID, &vlistID, 
                                     sizeof ( streamID ));
            modelWinBufferPutAtEnd ( __func__, collID, &tokenSep, 
                                     sizeof ( tokenSep ));
          }
        
        xdebug ( "WROTE FUNCTION CALL IN BUFFER OF WINS:  %s, streamID=%d,"
                 " vlistID=%d", 
                 funcMap[funcID], streamID, vlistID );
      }
      break;
    default:
      xabort ( "FUNCTION NOT MAPPED!" );
    }
  
  va_end ( ap );  
}

#endif

/*****************************************************************************/

int pioInit ( int comm, int nIOP, int mode )
{
#ifdef USE_MPI
#endif
  return 0;
}

/*****************************************************************************/
/** 
   @brief initializes the MPI_Communicators needed for the 
  communication between the calculator PEs and the I/O PEs and within the 
  group of I/O PEs.

  commGlob: all PEs

  commPIO: I/O PEs, PEs with highest ranks in commGlob

  commModel: calculating PEs, no I/O PEs

  commsIO[i]:
 
  Collective call 

  @param comm MPI_Communicator of all calling PEs
  @param nIOP number of I/O PEs
  @return int indicating wether the calling PE is a calcutator (1) or not (0)
*/

#ifdef USE_MPI
MPI_Comm pioInit_c ( MPI_Comm commGlob, int nProcsIO, int IOMode )
{
  int sizeGlob;

  if ( IOMode < PIO_MINIOMODUS || IOMode > PIO_MAXIOMODUS )
    xabort ( "PIOTYPE is no valid modus" );
  
#ifdef _SX
  if ( IOMode ==  PIO_POSIX_ASYNCH )
    xabort ( "PIO_POSIX_ASYNCH does not work on SX" );
#endif

  commInit ();
  commDefCommGlob ( commGlob );

  sizeGlob = commInqSizeGlob ();                                       
  if ( nProcsIO <= 0 || nProcsIO >= sizeGlob || nProcsIO > sizeGlob / 2 )
    xabort ( "bad distribution of tasks on PEs" );

  commDefNProcsIO ( nProcsIO );
  commDefIOMode   ( IOMode, PIO_MAXIOMODUS, PIO_MINIOMODUSWITHSPECIALPROCS );
  commDefCommPio  ();

  // just for test case compareResourceArray
  if ( commInqRankGlob () == 0 && commInqSizeGlob () == 1 ) 
    return commInqCommGlob ();

  if ( commInqIsProcIO ()) 
    { 
      IOServer ();
      commDestroy ();
      MPI_Finalize ();
      exit ( 0 );
    }
  else
    { 
      commEvalPhysNodes ();
      commDefCommsIO ();        
    }
  
  return commInqCommModel ();
}
#endif

/************************************************************************/

void  pioEndDef ( void )
{
#ifdef USE_MPI
  char   * buffer;
  int bufferSize;
  int rankGlob = commInqRankGlob ();

  varsMapNDeco ( commInqNNodes (), commInqNodeSizes ());

  if ( rankGlob < commInqNProcsColl ())
    {
      reshPackBufferCreate ( &buffer, &bufferSize, commInqCommsIO ( rankGlob ));
      
      xmpi ( MPI_Send ( buffer, bufferSize, MPI_PACKED, commInqNProcsModel (), 
                        RESOURCES, commInqCommsIO ( rankGlob )));
      
      xdebug ( "sent message RESOURCES" );

      namespaceDefResASent ();
      
      reshPackBufferDestroy ( &buffer );
    }

  modelWinCreate ();
#endif
}


/****************************************************/


/**
  @brief is invoked by the calculator PEs, to inform
  the I/O PEs that no more data will be written.

  @param

  @return
*/

void pioFinalize ()
{
#ifdef USE_MPI
  int collID, ibuffer = 1111;
  
  for ( collID = 0; collID < commInqNProcsColl (); collID++ )
    {
      xmpi ( MPI_Send ( &ibuffer, 1, MPI_INT, commInqNProcsModel (), 
                        FINALIZE, commInqCommsIO ( collID )));
      xdebug ( "sent message FINALIZE" );
    }

  modelWinCleanup ();
  commDestroy ();
#endif
} 
 
 /************************************************************************/
 
void pioWriteTimestep ( int ts, int vdate, int vtime )
{
#ifdef USE_MPI
  int collID, buffer[timestepSize], iAssert = 0;
  int tokenEnd = END;
  int rankGlob = commInqRankGlob ();
  int nProcsColl = commInqNProcsColl ();
  int nProcsModel = commInqNProcsModel ();

  if ( ts             < 0                    ||
       vdate          < 0                    ||
       vtime          < 0                    ||
       winPostSet    == NULL                 ||
       win           == NULL )
    xabort ( "" );

  buffer[0] = ts;
  buffer[1] = vdate;
  buffer[2] = vtime;
  
  if ( rankGlob < nProcsColl ) 
    {
      xmpi ( MPI_Send ( &buffer[0], timestepSize, MPI_INTEGER, nProcsModel, 
                        WRITETS, commInqCommsIO ( rankGlob )));
      xdebug ( "sent message WRITETS" );
    }

  for ( collID = 0; collID < nProcsColl; collID++ )
    {
      if ( winPostSet[collID] == 1 ) xabort ( "" );
      modelWinBufferPutAtEnd ( __func__, collID, &tokenEnd, sizeof ( tokenEnd ));
      xmpi ( MPI_Win_post ( groupsIONetto[collID], iAssert, win[collID] ));
      winPostSet[collID] = 1;
    } 
#endif
}


/************************************************************************/
/*
 * Local Variables:
 * c-file-style: "Java"
 * c-basic-offset: 2
 * indent-tabs-mode: nil
 * show-trailing-whitespace: t
 * require-trailing-newline: t
 * End:
 */
